/*
 * LED Toy, Control Program
 *
 * Written 2006 by Werner Almesberger
 * Copyright 2006 Werner Almesberger
 */


#include <stdint.h>
#include <string.h>
#include <setjmp.h>
#include <avr/io.h>
#include <avr/sleep.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>

#define F_CPU	1000000UL
#include <util/delay.h>


/*
 * PB0	-
 * PB1	-
 * PB2	-
 * PB3	MOSI
 * PB4	MISO
 * PB5	SCK
 * PB6	-
 * PB7	Push button (to GND)
 *
 * PC0	D21/D22
 * PC1	D19/D20
 * PC2	D17/D18
 * PC3	D15/D16
 * PC4	D13/D14
 * PC5	D11/D12
 * PC6	Reset
 *
 * PD0	D9/D10
 * PD1	D7/D8
 * PD2	D5/D6
 * PD3	D3/D4
 * PD4	D1/D2
 * PD5	Phase B
 * PD6	Phase A
 * PD7	D23/D24
 */

/*
 * Data encoding:
 *
 * Byte 0:
 *
 * 0	PC0, phase 0
 * 1	PC1, phase 0
 * 2	PC2, phase 0
 * 3	PC3, phase 0
 * 4	PC4, phase 0
 * 5	PC5, phase 0
 * 6	PC4, phase 1	>> 2
 * 7	PC5, phase 1	>> 2
 *
 * Byte 1:
 *
 * 0	PD0, phase 0
 * 1	PD1, phase 0
 * 2	PD2, phase 0
 * 3	PD3, phase 0
 * 4	PD4, phase 0
 * 5	PC2, phase 1	>> 3
 * 6	PC3, phase 1	>> 3
 * 7	PD7, phase 0
 *
 * Byte 2:
 *
 * 0	PD0, phase 1
 * 1	PD1, phase 1
 * 2	PD2, phase 1
 * 3	PD3, phase 1
 * 4	PD4, phase 1
 * 5	PC0, phase 1	>> 5
 * 6	PC1, phase 1	>> 5
 * 7	PD7, phase 1
 *
 * Note: phase 0 is phase "B", phase 1 is phase "A".
 */

#define	LEDS		24		/* Number of LEDs */

#define	SEQ_NEXT	0		/* Sequence flag: go to next image */
#define	SEQ_OFF		1		/* Power down */
#define	SEQ_LOOP	2		/* Loop back */

#define	GROUP_RUN	10		/* LED groups for the menu */
#define	GROUP_SELECT	14
#define GROUP_ROTATE	16
#define	GROUP_FREQ	18
#define	GROUP_SPACE	20
#define	GROUP_TIMEOUT	22

#define	T_IDLE_MS	5000		/* idle timer for menu, 5s */
#define	T_LONG_MS	400		/* long button press, 400ms */
#define	T_CONFIG_MS	30000		/* idle timer for configuraion, 30s */
#define	T_CURRENT_MS	250		/* show current setting, 250ms */

#define	W_BOUNCE_MS	30		/* debouncing delay, 30ms */
#define	W_PHASE_US	10		/* phase duration, 10us each */
#define	W_STARTUP_MS	50		/* set each LED for 50 ms at startup */

#define	BUTTON_UP	(PINB & 0x80)
#define	PHASE_A		(1 << 6)
#define	PHASE_B		(1 << 5)

#define	OFF_C		0x3f
#define	OFF_D		0xff


/* ----- Configuration variables ------------------------------------------- */


static unsigned char freq = 16;	/* Display frequency, index into freq_map */
static unsigned char timeout = 5; /* Timeout, in minutes. 0 = forever */
static unsigned char gap = 10;	/* Gap between images, in pixels */
static uint8_t rotation = 0;	/* Image rotation, in 90 degree turns */


/* ----- Image data -------------------------------------------------------- */


#define index_start __data_load_end
extern const prog_uint16_t __data_load_end;

static const prog_uint16_t *current_index = &index_start;
static const prog_uint8_t *image;

static uint8_t image_buffer[LEDS*3];


/* ----- Miscellaneous variables and items --------------------------------- */


static jmp_buf jmp_off;

/* Frequency conversion from "freq" to the number of loops. */

static const uint8_t freq_map[] = {
#include "freq.inc"
};


/* ----- Timers ------------------------------------------------------------ */


/*
 * We run a 1 MHz/256/256 = 15.26 Hz clock. This yields a resolution of
 * 65.5 ms and a maximum period of 2147 seconds or 35.8 minutes. The
 * maximum timer-based delay in the system is the display timeout of 23
 * minutes.
 */


typedef unsigned int TIMER;

static volatile TIMER current_ticks;
static TIMER now;


static void update_time(void)
{
    cli();
    now = current_ticks;
    sei();
}


#define set_timer_ticks(t,ticks) \
    ({ (t) = now+(ticks); })

#define set_timer_ms(t,ms) \
    set_timer_ticks(t,(ms) >> 6)


static int timer_expired(TIMER t)
{
    return (int) now-(int) t >= 0;
}


ISR(TIMER0_OVF_vect)
{
    current_ticks++;
}


/* ----- Button input ------------------------------------------------------ */


static int button_down;


static int button_changed(void)
{
    update_time();
    if (button_down == !BUTTON_UP)
	return 0;
    button_down = !button_down;
    _delay_ms(W_BOUNCE_MS);
    return 1;
}


ISR(SIG_PIN_CHANGE0)
{
    /* we need the interrupt handler only for wakeup */
}


/* ----- Individual LEDs --------------------------------------------------- */


static const uint8_t led_map[LEDS][2] = {
#include "led_map.inc"
};


static void led_off(void)
{
    PORTC = OFF_C;
    PORTD = OFF_D;
}


/*
 * If we have a uint8_t index, led_map[(int) index] or led_map[index] (with
 * "index" passed as "int") sometimes generate considerably more code than the
 * equivalent *(led_map+index).
 */


static void show_led(uint8_t num)
{
    PORTC = (*(led_map+num))[0];
    PORTD = (*(led_map+num))[1];
}


static void led_pair(uint8_t first)
{
    PORTC = (*(led_map+first))[0];
    PORTD = (*(led_map+first))[1] & ~(PHASE_A | PHASE_B);
}


/* ----- Image display (low-level) ----------------------------------------- */


static void show_frame(const uint8_t *start,uint8_t cols)
{
    const uint8_t *p;
    uint8_t c0,d0,c1,d1;
    uint8_t i,j;

    for (p = start; p != start+3*cols;) {
	if (cols == LEDS) {
	    c0 = *p++;
	    d0 = *p++;
	    d1 = *p++;
	}
	else {
	    c0 = pgm_read_byte(p++);
	    d0 = pgm_read_byte(p++);
	    d1 = pgm_read_byte(p++);
	}
	c1 = ((c0 & 0xc0) >> 2) | ((d0 & 0x60) >> 3) | ((d1 & 0x60) >> 5);
	d0 = (d0 | PHASE_B) & ~PHASE_A;
	d1 = (d1 | PHASE_A) & ~PHASE_B;
	for (j = freq_map[freq]; j; j--) {
	    PORTC = c0;
	    PORTD = d0;
	    _delay_us(W_PHASE_US);
	    PORTD = OFF_D;
	    PORTC = c1;
	    PORTD = d1;
	    _delay_us(W_PHASE_US);
	    PORTD = OFF_D;
	}
    }
    for (i = gap; i; i--)
	for (j = freq_map[freq]; j; j--)
	    _delay_us(W_PHASE_US*2+10); /* estimate 10 cycles overhead */
}


static void show_image(const prog_uint8_t *start)
{
    uint8_t cols;

    cols = pgm_read_byte(start);
    if (cols <= LEDS)
	show_frame(image_buffer,LEDS);
    else
	show_frame(start+2,cols);
}


static void set_image(const prog_uint16_t *addr)
{
    uint8_t cols;

    /* get object pointer from index */
    image = (prog_uint8_t *) pgm_read_word(addr);

    /* if object is a sequence, the title image is in the next word */
    if (!pgm_read_byte(image))
	image = (prog_uint8_t *) pgm_read_word(image+2);

    memset(image_buffer,0xff,LEDS*3);
    cols = pgm_read_byte(image);
    if (cols <= LEDS)
	memcpy_P(image_buffer+3*((LEDS-cols) >> 1),image+2,cols*3);
}


/* ----- Image rotation ---------------------------------------------------- */


static const uint8_t bit_pos[24] = {
#include "bit_pos.inc"
};


static uint8_t swap_pixel(uint8_t x,uint8_t y,uint8_t bit)
{
    uint8_t *p,set;
    uint8_t pix;

    pix = bit_pos[y];
    p = image_buffer+3*x+(pix >> 3);
    pix = 1 << (pix & 7);
    set = *p & pix;
    if (bit)
	*p |= pix;
    else
	*p &= ~pix;
    return set;
}


static void rotate_image(void)
{
    uint8_t x,y,tmp;

    for (x = 0; x != LEDS/2; x++)
	for (y = 0; y != LEDS/2; y++) {
	    tmp = swap_pixel(x,y,0);
	    tmp = swap_pixel(LEDS-1-y,x,tmp);
	    tmp = swap_pixel(LEDS-1-x,LEDS-1-y,tmp);
	    tmp = swap_pixel(y,LEDS-1-x,tmp);
	    swap_pixel(x,y,tmp);
	}
}


/* ----- User interface helper functions ----------------------------------- */


/*
 * We break things apart into many little functions, to save code space.
 */


#define UI_STATE(timeout,action,on_change,on_timeout) \
    do {					\
	set_timer_ms(t,timeout);		\
	while (1) {				\
	    action;				\
	    if (button_changed()) {		\
		on_change;			\
		break;				\
	    }					\
	    if (timer_expired(t)) {		\
		on_timeout;			\
		break;				\
	    }					\
	}					\
    }						\
    while(0)


static void __attribute__((__noreturn__)) idle(void)
{
    longjmp(jmp_off,1);
}


static void ui_settle(void)
{
    TIMER t;

    led_off();
    UI_STATE(T_IDLE_MS,,,idle());
}


static int ui_menu(void)
{
    TIMER t;

    UI_STATE(T_IDLE_MS,,,idle());
    UI_STATE(T_LONG_MS,,return 0,);
    ui_settle();
    return 1;
}


/* ----- Image selection and display (high-level) -------------------------- */


static const prog_uint8_t *obj;	/* Currently selected object */
static const prog_uint8_t *seq;	/* Current image entry in sequence, NULL if
				  object is not a sequence */
static TIMER seq_t;		/* Timeout to the next item in the sequence */



static int ui_image(int hold)
{
    TIMER t;

    if (hold)
	UI_STATE(T_CURRENT_MS,,goto rush,);
    UI_STATE(T_CONFIG_MS,show_image(image),,idle());
rush:
    UI_STATE(T_LONG_MS,show_image(image),return 0,);
    ui_settle();
    return 1;
}


static void enter_sequence(void)
{
    uint8_t i;

    set_image((prog_uint16_t *) (seq+4));
    for (i = (pgm_read_byte(seq)+rotation) & 3; i; i--)
	rotate_image();
    set_timer_ticks(seq_t,pgm_read_word(seq+2));
}


static void leave_sequence(void)
{
    /*
     * We could save 6 bytes by turning this into an "if" plus an assignment
     * with the ternary operator. However, the "switch" is much more readable
     * and we're no *that* desperate for space.
     */
    switch (pgm_read_byte(seq+1)) {
	case SEQ_LOOP:
	    seq = obj+4;
	    break;
	case SEQ_OFF:
	    idle();
	    /* does not return */
	default:
	    seq += 6;
    }
}


static void reset_sequence(void)
{
    obj = (prog_uint8_t *) pgm_read_word(current_index);
    if (pgm_read_byte(obj))
	seq = NULL;
    else {
	seq = obj+4;
	enter_sequence();
    }
}


static void ui_run_image(void)
{
    TIMER t;

    if (seq && timer_expired(seq_t)) {
	leave_sequence();
	enter_sequence();
    }
    show_image(image);
    if (!button_changed())
	return;
    UI_STATE(T_LONG_MS,show_image(image),return,);
    ui_settle();
    idle();
}


/* ----- The menu ---------------------------------------------------------- */


static void menu(void)
{
    TIMER t;
    uint8_t i;

    set_image(current_index);
    for (i = rotation & 3; i; i--)
	rotate_image();

    led_pair(GROUP_RUN);
    UI_STATE(T_LONG_MS,,,return);

ui_run_ind:
    led_pair(GROUP_RUN);
    if (ui_menu()) {
	reset_sequence();
	if (!timeout) 
	    while (1)
		ui_run_image();
	else {
	    set_timer_ticks(t,timeout*(60000/65));
	    while (!timer_expired(t))
		ui_run_image();
	}
	return;
    }

    while (led_pair(GROUP_SELECT), ui_menu())
	while (1) {
	    rotation = 0;
	    set_image(current_index);
	    if (ui_image(0))
		break;
	    current_index++;
	    if (!pgm_read_word(current_index))
		current_index = &index_start;
	}

    while (led_pair(GROUP_ROTATE), ui_menu())
	while (1) {
	    if (ui_image(0))
		break;
	    rotate_image();
	    rotation++;
	}


    while (led_pair(GROUP_FREQ), ui_menu())
	while (1) {
	    show_led(freq);
	    if (ui_image(1))
		    break;
	    if (freq)
		freq--;
	    else
		freq = LEDS-1;
	}

    while (led_pair(GROUP_SPACE), ui_menu())
	while (1) {
	    show_led(gap);
	    if (ui_image(1))
		break;
	    if (gap)
		gap--;
	    else
		gap = LEDS-1;
	}

ui_timeout_ind:
    led_pair(GROUP_TIMEOUT);
    if (ui_menu())
	while (1) {
	    show_led(timeout);
	    UI_STATE(T_CONFIG_MS,,,return);
	    UI_STATE(T_LONG_MS,,,({ ui_settle(); goto ui_timeout_ind; }));
	    if (timeout)
		timeout--;
	    else
		timeout = LEDS-1;
	}

    goto ui_run_ind;
}


/* ----- Main loop and setup ----------------------------------------------- */


int main(void)
{
    int i;

    /* Configure port B as input, activate pull-up on PB7 */
    DDRB = 0;
    PORTB = 0x80;

    /* Configure port C[0-6] and D as output, default to "off" state (H) */
    led_off();
    DDRC = 0x3f;
    DDRD = 0xff;

    for (i = 0; i != LEDS; i++) {
	show_led(i);
	_delay_ms(W_STARTUP_MS);
    }

    TCCR0B = 4;	/* run timer 0 at clk/256, yielding a 15.26 Hz clock */
    TIMSK0 = 1 << TOIE0; /* enable timer interrupts */

    PCMSK0 = 0x80; /* enable PCINT7 aka PB7 */
    PCICR = 1;

    sei();

    while (1) {
	led_off();

	do {
	    set_sleep_mode(SLEEP_MODE_PWR_DOWN);
	    sleep_mode();
	    PCIFR = 0; /* clear interrupt */
	}
	while (BUTTON_UP);
	button_down = 1;
	_delay_ms(W_BOUNCE_MS);

	if (!setjmp(jmp_off))
	    menu();
    }
}
