#!/usr/bin/perl
#
# lta24.pl - Merge LED Toy firmware with images
#
# Written 2005, 2006 by Werner Almesberger
# Copyright 2005, 2006 by Werner Almesberger
#

#
# ATmega48 memory
#
$MAX_BYTES = 4096;

#
# Timing
#
$T = 1000000/256/256;	# timer cycle

#
# Map of LED position to bit in phase/port
#
# Bytes are arranged as follows:
#
# Phase A, port C	(bits 0 to 7)
# Phase A, port D	(bits 8 to 15)
# Phase B, port C	(bits 16 to 23)
# Phase B, port D	(bits 24 to 31)
#

@led_map =
  ( 7+24, 7+ 8, 0+16, 0,    1+16, 1,    2+16, 2,
    3+16, 3,    4+16, 4,    5+16, 5,    0+24, 0+ 8,
    1+24, 1+ 8, 2+24, 2+ 8, 3+24, 3+ 8, 4+24, 4+ 8 );

#
# Map of image bit positions to port bits
#
# Ports are arranged as follows:
#
# Phase 0, port C	(bits 0 to 7)
# Phase 0, port D	(bits 8 to 15)
# Phase 1, port C	(bits 16 to 23)
# Phase 1, port D	(bits 24 to 31)
#
#

@image_to_port =
  ( 0+ 0, 1+ 0, 2+ 0, 3+ 0, 4+ 0, 5+ 0, 4+16, 5+16,
    0+ 8, 1+ 8, 2+ 8, 3+ 8, 4+ 8, 2+16, 3+16, 7+ 8,
    0+24, 1+24, 2+24, 3+24, 4+24, 0+16, 1+16, 7+24 );

#
# XOR mask over the values to output
#

$xor = 0xffffffff;


sub convert
{
    local $width = shift(@_);
    local ($x,$y);
    local @res = ();
    local $bit;

    for ($x = 0; $x != $width; $x++) {
	$pat = 0;
	for ($y = 0; $y != 24; $y++) {
	    $pix = $x+$y*(($width+7) & ~7);
	    if ($_[$pix >> 3] & (1 << ($pix & 7))) {
		$bit = $port_to_image[$led_map[$y]];
		die "inaccessible port bit" unless defined $bit;
		$pat |= 1 << $bit;
	    }
	}
	push(@res,$pat);
    }
    return @res;
}


sub load_xbm
{
    local $file = $_[0];
    local ($s,$x,$bytes,@res);

    if ($file !~ /^\// && ! -r $file) {
	foreach (@search) {
	    if (-r $_."/".$file) {
		$file = $_."/".$file;
		last;
	    }
	}
    }
    open(FILE,$file) || die "$file: $!";
    $s = join("",<FILE>);
    close FILE;
    $s =~ /#define\s\S+_width\s(\d+)\s+#define\s\S+_height\s24\s+/ ||
      die "$file: not a valid X bitmap file";
    die "image width $1 out of range [1, 255]" if $1 < 1 || $1 > 255;
    $x = $1;
    $bytes = 3*(($x+7) & ~7)-1;
    $' =~ /^{0}static\s+(unsigned\s+)*char\s+\S+_bits\[\]\s=\s\{\s+
{0}(0x[0-9a-f]{2}(,\s*0x[0-9a-f]{2}){$bytes})\};\s*$/ ||
      die "$file: not a valid X bitmap file";
    return ($x,map { hex $_; } split(/,\s*/,$2));
}


sub load_img
{
    local $spec = $_[0];
    local ($x,$i,$res);
    local @data;
    local $flyback = 0;

    if ($spec =~ m|\s*//\s*(\d+)\s*$|) {
	$flyback = $1;
	$spec = $`;
    }
    while (length $spec) {
	local ($op,$not,$cx,@cd);

        $spec =~ /^\s*([|&^])\s*|/;
	($op,$spec) = ($1,$');
	if ($op && !defined $x) {
	    die "image spec must not begin with an operator \"$&\"";
	}
	if (!$op && defined $x) {
	    die "file names not joined by an operator \"$&\"";
	}
	$spec =~ /^\s*(~?)\s*([^&|]*[^&| \t\n])\s*([|&^]?.*|$)/ ||
	  die "garbled image spec \"$spec\"";
	$not = $1 eq "~";
	$spec = $3;
	($cx,@cd) = &load_xbm($2);
	map { $_ = ~$_ } @cd if $not;
	if (!$op) {
	    $x = $cx;
	    @data = @cd;
	}
	else {
	    local $i;

	    die "cannot combine images with different sizes ($cx != $x)"
	      if $x != $cx;
	    for ($i = 0; $i != @data; $i++) {
		$data[$i] = $op eq "&" ? $data[$i] & $cd[$i] :
		  $op eq "|" ? $data[$i] | $cd[$i] : $data[$i] ^ $cd[$i];
	    }
	}
    }
    @data = &convert($x,@data);
    if ($x == 24) {	# for historical reasons, we only trim 24x24 images
	local ($left,$right);

	for ($left = 0; $left != $x; $left++) {
	    last if $data[$left];
	}
	for ($i = $left; $i != $x; $i++) {
	    $right = $i if $data[$i];
	}
	for ($i = 0; $i != $left; $i++) {
	    shift(@data);
	}
	$x = $right-$left+1;
    }
    print STDERR "warning: flyback ($flyback) is less than image width ($x)\n"
      if $flyback && $flyback < $x;
    die "flyback value $flyback is too large (> 1008)" if $flyback > 1008;
    $res = pack("CC",$x,($flyback+15) >> 4);
    for ($i = 0; $i != $x; $i++) {
	local $tmp = $data[$i] ^ $xor;
	$res .= pack("CCC",$tmp,$tmp >> 8,$tmp >> 16);
    }
    return $obj{$res} if defined $obj{$res};
    $obj{$res} = pack("v",2*@ARGV+2+length($data)+length($bin));
    $data .= $res;
    return $obj{$res};
}


sub load_seq
{
    local $file = $_[0];
    local ($s,$res,$end,$retval);

    open(FILE,$file) || die "$file: $!";
    $s = join("",<FILE>);
    close FILE;
    $s =~ s/\s*#[^\n]*\n+/\n/g;
    $s =~ /title\s+([^\n]*?\S)\s*?\n/ || die "$file: no title object";
    $s = $';
    $res = pack("v",0x200).&load_img($1);
    $end = 0;
    while (length $s) {
	local $flags;

	die "$file: data beyond the end" if $end;
	$s =~ /^image\s+([^\n]*?\S)\s*rotate\s+([0123])\s*
{0}timeout\s+(\d+(\.\d*)?)\s*(loop|end)?\s*/ ||
	  die "$file: invalid object entry";
	$s = $';
	$end = $5 ne "";
	$flags = 0;
	$flags |= 1 if $5 eq "end";
	$flags |= 2 if $5 eq "loop";
	$res .= pack("CC",$2,$flags);
	$res .= pack("v",$3*$T);
	$res .= &load_img($1);
    }
    die "$file: no end marker" if !$end;
    $retval = pack("v",2*@ARGV+2+length($data)+length($bin));
    $data .= $res;
    return $retval;
}


sub usage
{
    print STDERR
"usage: $0 [-I dir:dir:...] [-o outfile] firmware.bin [display-item ...]\n".
"       $0 [-I dir:dir:...] -C firmware.bin image-expression\n".
"       $0 -M\n".
"       $0 -B\n\n".
"  -I dir:dir:...  add the directories to the search path for images\n".
"  -o outfile      write the resulting binary to the file instad of stdout\n".
"  -C              generate a C intializer for one image instead of compound\n".
"                  firmware binary. Note that only the name of firmware.bin is".
"                  used but the file is not actually read.\n".
"  -M and -B       are used in the built process. See Makefile for details.".
"\n\n".
"  display-expression: image-expression | file.seq\n".
"  image-expression: image-spec [ binary-operator image-spec ... ] [ flyback ]\n".
"  image-spec: [ unary-operator] image.xbm\n".
"  unary-operator: ~\n".
"  binary-operator: &, |, or ^\n".
"  flyback: // number-of-pixels\n";
    exit(1);
}


while ($ARGV[0] eq "-o" || $ARGV[0] eq "-I") {
    if ($ARGV[0] eq "-o") {
	&usage if $#ARGV < 1;
	&usage if defined $output;
	shift(@ARGV);
	$output = shift(@ARGV);
    }
    else {
	&usage if $#ARGV < 1;
	&usage if defined $output;
	shift(@ARGV);
	push(@search,split(":",shift(@ARGV)));
    }
}

if ($ARGV[0] eq "-C") {
    &usage unless $#ARGV == 2;
    shift(@ARGV);
    $c = 1;
}
elsif ($ARGV[0] eq "-M") {
    &usage unless $#ARGV == 0;
    $n = 0;
    print "/* MACHINE-GENERATED. DO NOT EDIT ! */\n\n";
    for (@led_map) {
	print sprintf("    { 0xff^0x%02x, 0xff^0x%02x },\t/* LED $n */\n",
	    $_ & 8 ? 0 : 1 << ($_ & 7),
	    ($_ & 8 ? 1 << ($_ & 7) : 0) |
	    ($_ & 16 ? 1 << 5 : 1 << 6));
	$n++;
    }
    exit;
}
elsif ($ARGV[0] eq "-B") {
    &usage unless $#ARGV == 0;
    $n = 0;
    print "/* MACHINE-GENERATED. DO NOT EDIT ! */\n\n";
    for (@image_to_port) {
	print sprintf(
	  "    %d+%2d,\t/* image byte, %d bit %d -> ".
	  "phase %d, port %s, bit %d */\n",
	    $_ & 7,$_ & ~7,$n >> 3,$n & 7,
	    $_ >> 4,$_ & 8 ? "D" : "C",$_ & 7);
	$n++;
    }
    exit;
}
$fw = shift(@ARGV);
&usage unless defined $fw;

for ($n = 0; $n != @image_to_port; $n++) {
    $port_to_image[$image_to_port[$n]] = $n;
}

if ($fw =~ /-LTA24V0\.bin$/) {
    # Nothing to do
}
else {
    print STDERR
      "warning: unknown firmware version, assuming V0 compatibility\n"
      if $fw !~ /-LTA24V[0]\.bin$/;
}

if (!$c) {
    open(FILE,$fw) || die "$fw: $!";
    while (<FILE>) {
	$bin .= $_;
    }
    close FILE;
}

$table = "";
$data = "";
for (@ARGV) {
    if ($_ =~ m|\.xbm(\s*//\s*\d+)?$|) {
	$table .= &load_img($_);
    }
    elsif ($_ =~ /\.seq$/) {
	usage if $c;
	$table .= &load_seq($_);
    }
    else {
	usage if $c;
	die "$_ is not a bitmap (.xbm) or sequence (.seq) file";
    }
}

if ($c) {
    while (length $data) {
	print sprintf(" 0x%02x,",unpack("C",substr($data,0,1)));
	$data = substr($data,1);
    }
    print "\n";
    exit(0);
}

($len_fw,$len_data) = (length $bin,2+length $data);
$bin .= $table.pack("v",0).$data;

$len = length($bin);
if ($len > $MAX_BYTES) {
    print STDERR "output is too big ($len > $MAX_BYTES)\n";
    exit(1);
}

if (defined $output) {
    open(FILE,">$output") || die "$output: $!";
    if (print FILE $bin) {
	if (!close FILE) {
	    $error = $!;
	    unlink($output);
	    die "$output: $error";
	}
    }
    else {
	$error = $!;
	unlink($output);
	die "$output: $error";
    }
}
else {
    $| = 1;
    die unless print $bin;
}

print STDERR $len,"/",$MAX_BYTES," bytes (",int(100*$len/$MAX_BYTES+0.5),
  "%): firmware ",$len_fw," (",int(100*$len_fw/$MAX_BYTES+0.5),
  "%), data ",$len_data," (",int(100*$len_data/$MAX_BYTES+0.5),"%)\n";
